/**************************************************************************************************

Copyright (c) 2007-2015   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mitaka.h"
#include  "mtk_3D_model.h"
#include  "mtk_minor_planets.h"
#include  "../common/astro/astro_const.h"
#include  "../common/astro/astrometry.h"
#include  "mtk_draw_common.h"
#include  "mtk_object.h"
#include  "mtk_objects.h"
#include  "mtk_target.h"
#include  "mtk_orbital_element.h"
#include  "mtk_solar_system.h"
#include  "mtk_sun.h"
#include  "mtk_planets.h"
#include  "mtk_label.h"
#include  "mtk_fixed_object.h"
#include  "mtk_space_curve.h"
#include  "../common/graphics/gl_Model3D.h"
#include  "mtk_calc_alpha.h"
#include  "mtk_switches.h"
#include  "mtk_colors.h"
#include  "mtk_gl_common.h"
#include  "mtk_OpenGL.h"
#include  "mtk_file.h"
#include  "mtk_dialog_box.h"
#include  "mtk_directory.h"
#include  <vector>
#include  <map>




static float  DwarfPlanetNameSize, SmallObjectNameSize;

// textures
static GLuint texCeres;
const char *fn_tex_Ceres = "Ceres.jpg";
// 3D models
static int  idxItokawaModel = -1;




//------------------
//  fт̃x
//------------------
const bilingual bszMinorPlanetLabel = { "Asteroid Belt", "{f|傤킭}{|}" };
const vector3d  minor_planet_label_pos( 2.5*_1AU_pc, 0, 0 );



static std::vector<small_object>	vSmallObjects;


//-------------------------------------------
//  fԍƏV̂̃CfbNX̎Qƃe[u
//-------------------------------------------
static map<long, long>		minor_planet_number_table;		


//----------------------------
//  Small objects Phys param
//----------------------------
struct  small_object_info
{
	int mp_no;				// minor planet number
	double Rx, Ry, Rz;		// size
	double B_V;				// B_V color
};

static small_object_info vSmoi[] = {
	{      1,  487.3,  487.3,  454, 1.0  },	//  Ceres [1]
	{      2,  260,  260,  260, 1.0  },		//  Pallas
	{      3,  120,  120,  120, 1.0  },		//  Juno
	{      4,  289,  280,  229, 1.0  },		//  Vesta [1]
	{    243, 26.8, 12.0,  7.6, 1.0  },		//  Ida [1]
	{    253,   33,   24,   23, 1.0  },		//  Mathilde
	{    433,   17.0, 5.5,  5.5, 1.0  },	//  Eros [1]
	{    951,  9.1, 5.2, 4.4, 1.0  },		//  Gaspra [1]
	{   1566,  0.5,  0.5,  0.5, 1.0  },		//  Icarus
	{  25143, 0.535, 0.294, 0.209, 1.0 },	//  Itokawa [1]
	{ 136199, 1200, 1200, 1200, 0.71 },		//  Eris
	{ 136108, 1000,  600,  500, 0.63 },		//  Haumea
	{  90377,  900,  900,  900, 1.23 },		//  Sedna
	{ 136472,  750,  750,  750, 0.9  },		//  Makemake
	{  90482,  500,  500,  500, 0.68 },		//  Orcus
	{  50000,  500,  500,  500, 0.94 },		//  Quaoar
	{  20000,  300,  300,  300, 0.85 }		//  Varuna
};
const int smo_phys_info_num = sizeof(vSmoi)/sizeof(vSmoi[0]);



//------------
//  3D f
//------------
bool
load_minorplanet_3D_model(const char *fn)
{
	//-- Load models
	idxItokawaModel = load_3D_model_file(DIR_DATA, fn);

	if (idxItokawaModel != -1) {
		Model3D *mdl = get_3DModel(idxItokawaModel);
		if (mdl != 0) { mdl->fit_size(2.0); }
		return true;
	}

	return false;
}



// ǂݍ
int
load_small_objects(FILE* fp, int type)
{
	while (true) {
		long	no;
		double	epoch;
		float	a, e, M0, n, node, i, omg;

		//--- read data
		int count;
		count = fread(&no,	sizeof(no),			1, fp);
		if (count < 1)	break;
		fread(&epoch,	sizeof(epoch),		1, fp);
		fread(&M0,		sizeof(M0),			1, fp);
		fread(&omg,		sizeof(omg),		1, fp);
		fread(&node,	sizeof(node),		1, fp);
		fread(&i,		sizeof(i),			1, fp);
		fread(&e,		sizeof(e),			1, fp);
		fread(&n,		sizeof(n),			1, fp);
		//
		a = pow(0.985607669/n, 2.0/3.0);
		//
		small_object  smo;
		smo.type			= type;
		smo.no				= no;
		smo.obj_idx			= -1;
		smo.label_obj_idx	= -1;
		smo.phys_param_idx	= -1;
		smo.bDwarfPlanet	= false;
		smo.bSelected		= false;
		smo.oe.set_parameters(epoch, a, e, M0, n, node, i, omg);

		//--  IuWFNgo^
		int idx = vSmallObjects.size();
		char key[256];
		sprintf(key, "SmallBody%d", smo.no);
		smo.obj_idx = register_object( mtk_object(OBJ_TYPE_SMALL_OBJECT, idx, 0.0f, true), key );

		vSmallObjects.push_back(smo);


		//--  fԍƂ̎Qƕ\  --
		minor_planet_number_table.insert( pair<long,long>(no, idx) );
	}


	//---  ufсṽx̓o^
	int fixed_obj_idx = add_fixed_object( minor_planet_label_pos, OBJ_TYPE_SMALL_OBJECT );
	mtk_object obj = mtk_object(OBJ_TYPE_FIXED_OBJECT, fixed_obj_idx, 0.0f, false);		//************
	int obj_idx = register_object(obj, "MinorPlanetsLabel0");

	bilingual  name = bszMinorPlanetLabel;
	float thc = 10 * D2R;
	add_label( name.text_J, name.text_E, obj_idx, MTK_COLOR_NAME_OBJECT, 1.0f, get_object_name_size(), 0, thc, false, "MinorPlanetsLabel" );
	

	//---  `̓o^
	for (int i=0; i<smo_phys_info_num; i++) {
		int idx = get_small_object_index( vSmoi[i].mp_no );
		if (idx == -1)  continue;

		small_object&  smo = vSmallObjects[idx];
		smo.phys_param_idx = i;
	}


	//--- 3Df
	load_minorplanet_3D_model("Itokawa.3ds");
	//load_minorplanet_3D_model("Itokawa.obj");

	return  vSmallObjects.size();
}



//-------------------
//  ] Ref.[1]
//-------------------
void
calc_minor_planet_orientation_angle(int mp_no, const double& t, double& alp, double& dlt, double& W)
{
	const double  JD0 = 2451545.0;	//  2000/1/1.5
	const double  d = t - JD0;

	switch (mp_no) {
	case 1:	// Ceres
	{
		alp = 291.0;
		dlt = 59.0;
		W = 170.90 + 952.1532 * d;
	}
	break;

	case 25143:	// Itokawa
	{
		alp = 90.53;
		dlt = -66.30;	// retrograde
		W = 0.0 + 712.143 * d;
	}
	break;

	default:
		alp = W = 0.0;
		dlt = 90.0;
		break;
	}
}


matrix3d
calc_minor_planet_ec_to_sat_matrix(int mp_no, const double& t)
{
	static matrix3d  M_J2000;
	static bool first = true;
	if (first) {
		first = false;
		M_J2000 = Rx3deg(epsJ2000);
	}

	// f̉]
	double alp, dlt, W;
	calc_minor_planet_orientation_angle(mp_no, t, alp, dlt, W);

	return  (Rzxz3deg(-W, -90 + dlt, -90 - alp) * M_J2000).transposed();
}

void
set_minor_planet_orientation(int mp_no, const double &t, const double& W_shift)
{
	set_object_orientation(calc_minor_planet_ec_to_sat_matrix(mp_no, t));	//***
	glRotated(W_shift, 0, 0, 1);
}




//----------------
//  O̓ǂݍ
//----------------
void
load_small_object_names(FILE* fp)
{
	const int BUF_SIZE = 80;
	char buf[BUF_SIZE], buf2[BUF_SIZE];
	char name_E[20], name_J[40];

	while (true) {
		int no;
		
		//-- Read one line
		if ( !read_one_line(fp, buf, BUF_SIZE) )  break;
		//-- Minor planet number
		if ( !extract_string(buf, 0, 6, buf2) ) continue;
		sscanf(buf2, "%ld", &no);
		//-- Name (English / Japanese)
		bool res_E, res_J;
		res_E = extract_string(buf, 10, 28, name_E);
		res_J = extract_string(buf, 30, 50, name_J);
		if (!res_E)		continue;
		if (!res_J)		strcpy(name_J, name_E);
		//--  Minor planet index
		int  idx = get_small_object_index( no );
		if (idx < 0)  continue;
		//-- Flag
		small_object&  smo = vSmallObjects[idx];
		smo.bDwarfPlanet = (buf[8] == 'D');
		smo.bSelected    = (buf[9] != ' ');


		//--  Register label
		float R   = 1.4 * 1000 * _1km_pc;

		float thc = 10 * D2R;
		int color_no;
		float name_size;
		if (smo.bDwarfPlanet) {
			color_no  = MTK_COLOR_NAME_DWARF_PLANET;
			name_size = DwarfPlanetNameSize;
		} else {
			color_no  = MTK_COLOR_NAME_SMALL_OBJECT;
			name_size = SmallObjectNameSize;
		}
				
		int obj_idx = vSmallObjects[idx].obj_idx;
		if (obj_idx == -1)  continue;

		char key[256];
		sprintf(key, "SmallBodyLabel%d", idx);
		smo.label_obj_idx = add_label( name_J, name_E, obj_idx, color_no, 1.0f, name_size, R, thc, true, key );
	}
}



//**********  x̓ǂݍ݌ɌĂ   ******
void
register_small_objects_orbit_curves()
{
	const int N = vSmallObjects.size();

	for (int i=0; i<N; i++) {
		const small_object&  smo = vSmallObjects[i];

		if ( smo.label_obj_idx == -1 )  continue;	// xׂ̂

		int color_no = (smo.bDwarfPlanet)  ?  MTK_COLOR_ORBIT_DWARF_PLANET  :  MTK_COLOR_ORBIT_SMALL_OBJECT;
		space_curve  sc( mtk_object(OBJ_TYPE_SMALL_OBJECT, i, 0.0f, false), get_orbit_width(), color_no );
		register_space_curve( sc );
	}
}


void
make_small_objects_orbit_curve(space_curve& sc, int div_N)
{
	const small_object&  smo = vSmallObjects[sc.get_related_object().get_index()];

	sc.clear();
	for (int j=0; j<div_N; j++) {
		double phi = 2.0 * M_PI * double(j) / div_N;
		sc.add_point( double(_1AU_pc) * smo.oe.calc_orbit_pos(phi) );
	}
}




//-------------
//  eNX`
//-------------
void
load_minor_planet_textures()
{
	SetSplashText("f̃eNX`Ǎ...", "Loading textures of minor planets...");

	texCeres = LoadTexture(fn_tex_Ceres, TexShrinkRate, false);
}





//---------
//  Alpha
//---------
float
get_small_object_label_alpha(const mtk_draw_struct& ds, int idx)
{
	const int type = get_small_object_type(idx);
	const float dscl = (type == SMALLOBJ_TYPE_MINOR_PLANETS) ? 3.0 * _1AU_pc : 60.0 * _1AU_pc;
	float  alp;
	
	if (norm( get_small_object_position(idx) - ds.tgt_pos ) > 10000*_1km_pc) {
		alp = calc_alpha_by_disapear_scale_inside( ds.scale, dscl );
	} else {
		alp = 1;
	}

	const vector3d  rpos = get_small_object_position( idx ) - ds.camera.get_pos();
	float  Rv = calc_visible_radius_small_object( idx );
	float  Dth = 2.0f * Rv / rpos.norm();

	return  alp * calc_alpha_by_apparent_diameter( Dth );
}





float
get_small_object_orbit_alpha(const mtk_draw_struct& ds, int idx)
{
	const int type = get_small_object_type(idx);
	const float dscl = (type == SMALLOBJ_TYPE_MINOR_PLANETS) ? 3.0 * _1AU_pc : 60.0 * _1AU_pc;
	float  alp;
	
	if (norm( get_small_object_position(idx) - ds.tgt_pos ) > 10000*_1km_pc) {
		alp = calc_alpha_by_disapear_scale_inside( ds.scale, dscl );
	} else {
		alp = calc_alpha_by_disapear_scale_inside( ds.scale, 1000000*_1km_pc );
	}

	const vector3d  rpos = get_small_object_position( idx ) - ds.camera.get_pos();
	float  Rv = calc_visible_radius_small_object( idx );
	float  Dth = 2.0f * Rv / rpos.norm();

	return  alp * calc_alpha_by_apparent_diameter( Dth );
}


//---------
//  index
//---------
int
get_small_object_index(int minor_planet_number)
{
	map<long, long>::iterator it = minor_planet_number_table.find( minor_planet_number );
	if (it==minor_planet_number_table.end())  return -1;	// no match
	return  it->second;
}

int
get_small_objects_num()
{
	return  vSmallObjects.size();
}


small_object&
get_small_object(int idx)
{
	return vSmallObjects[idx];
}




void
update_small_objects_position(const double& t)
{
	for (std::vector<small_object>::iterator it=vSmallObjects.begin(); it != vSmallObjects.end(); it++) {
		it->oe.update_pos_and_velocity(t);
	}
}


int
get_small_object_type(int idx)
{
	return  vSmallObjects[idx].type;
}

bool
is_small_object_selected(int idx)
{
	return  vSmallObjects[idx].bSelected;
}


void
get_small_object_size(int idx, double& Rx, double& Ry, double& Rz)
{
	int  pi_idx = vSmallObjects[idx].phys_param_idx;
	if (pi_idx == -1)  {
		Rx = Ry = Rz = 0.0;
		return;
	}

	Rx = vSmoi[pi_idx].Rx;
	Ry = vSmoi[pi_idx].Ry;
	Rz = vSmoi[pi_idx].Rz;
}


//-------------
//  Text size
//-------------
float
get_dwarf_planet_name_size()
{
	return  DwarfPlanetNameSize;
}

float
get_small_object_name_size()
{
	return  SmallObjectNameSize;
}

void
set_dwarf_planet_name_size(float sz)
{
	DwarfPlanetNameSize = sz;
}

void
set_small_object_name_size(float sz)
{
	SmallObjectNameSize = sz;
}



//------------------
//  Landing radius
//------------------
double
get_landing_radius_small_object(int idx)
{
	double Rx, Ry, Rz, R;
	get_small_object_size(idx, Rx, Ry, Rz);

	if (Rx > 0.0) {
		R = Rx * _1km_pc;
	} else {
		R = 100 * _1km_pc;	//***
	}

	return  R;
}




//----------------
//  Prepare draw
//----------------
bool
prepare_draw_small_object(const mtk_draw_struct& ds, int idx, double* pRb, double* pRh)
{
	//---  Visible
	int ty = get_small_object_type(idx);

	bool bFlg;
	switch (ty) {
		case SMALLOBJ_TYPE_MINOR_PLANETS:
			bFlg = test_disp_switch( DISPSW_MINOR_PLANET,	DISPM_DISP);	break;
		case SMALLOBJ_TYPE_TNOS:
			bFlg = test_disp_switch( DISPSW_TNO,	DISPM_DISP);			break;
		default:
			return  false;
	}
	if (!bFlg)  return false;






	//---  Radius
	double Rx, Ry, Rz;
	get_small_object_size(idx, Rx, Ry, Rz);
	if (Rx > 0.0) {
		*pRh = Rx * _1km_pc;
		*pRb = 1.5 * (*pRh);
	} else {
		*pRb = *pRh = 0.01 * _1km_pc;	//***
	}

	return  true;
}


float
get_minor_planet_point_alpha(const mtk_draw_struct& ds, int idx)
{
	const orbital_elements& oe = vSmallObjects[idx].oe;

	bool bTgt = (get_target().get_type() == OBJ_TYPE_SMALL_OBJECT && get_target().get_index()==idx);

	// V̂̈ʒu
	const vector3d  pos = double(_1AU_pc) * oe.get_pos();

	const float R1 = 0.01;
	const float R2 = 0.04;
	const float R3 = 1;
	const float R4 = 4;
	float R = norm(pos) / ds.scale;
	if (!bTgt && R > R4)  return  0.0;
	if (R < R1)  return  0.0;


	float af = 1.0;
	float sz = 1.5f / (ds.rad_per_pixel / 0.001);
	if (sz < 1) {
		af *= sz;
		sz = 1;
	}

	if (!bTgt && R > R3)	{ af *= 1 - (R-R3)/(R4-R3); }
	if (R < R2)				{ af *= (R-R1)/(R2-R1); }
	
	
	if (af < 0.05)  return 0.0;
	return  af;
}



//------------
//  Position
//------------
vector3d
get_small_object_position(int idx)
{
	return double(_1AU_pc) * vSmallObjects[idx].oe.get_pos();
}


//---------
//  Scale
//---------
double
get_scale_small_object(int idx)
{
	double Rx, Ry, Rz, R;
	get_small_object_size(idx, Rx, Ry, Rz);

	if (Rx > 0.0) {
		R = 30 * Rx * _1km_pc;
	} else {
		R = 30 * 100 * _1km_pc;
	}

	return  R;
}



//---------
//  Label
//---------
bool
is_small_object_label_visible(int idx)
{
	int type = get_small_object_type( idx );

	int dispsw_idx;
	if (type == SMALLOBJ_TYPE_TNOS)		dispsw_idx = DISPSW_TNO;
	else								dispsw_idx = DISPSW_MINOR_PLANET;

	if ( !test_disp_switch( dispsw_idx, DISPM_DISP) || !test_disp_switch( dispsw_idx, DISPM_NAME ) )  return  false;

	if (test_disp_switch(dispsw_idx, DISPM_SELECT) && !is_small_object_selected(idx))  return  false;

	return  true;
}







//--------
//  Draw
//--------
void
setup_drawing_small_object(int idx)
{
	int  pi_idx = vSmallObjects[idx].phys_param_idx;
	if (pi_idx == -1)  return;

	//--  material
	RGBf  col = correct_RGB( blackbody_color( B_V_to_Tc( vSmoi[pi_idx].B_V ) ) );
	float mat[4];
	mat[0] = col.r;
	mat[1] = col.g;
	mat[2] = col.b;
	mat[3] = 1;
	glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat );
	mat[0] *= 0.5;
	mat[1] *= 0.5;
	mat[2] *= 0.5;
	glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat );

	glEnable( GL_LIGHTING );
	glEnable( GL_LIGHT0 );
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);


	if (vSmallObjects[idx].no==1 &&  texCeres != -1) {
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texCeres);
	}
	else {
		glDisable(GL_TEXTURE_2D);
	}
}

void
draw_a_small_object(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const double& scl)
{
	int  pi_idx = vSmallObjects[idx].phys_param_idx;
	if (pi_idx == -1)  return;

	vector3d  pos = get_small_object_position( idx );
	double Rx, Ry, Rz;
	Rx = vSmoi[pi_idx].Rx * scl * _1km_pc;
	Ry = vSmoi[pi_idx].Ry * scl * _1km_pc;
	Rz = vSmoi[pi_idx].Rz * scl * _1km_pc;

	glPushMatrix();
		glTranslatef(rpos.x, rpos.y, rpos.z);


		if (!bSideLight)	set_light_from_sun(pos);
		else				set_side_light(ds);

		const int mp_no = vSmallObjects[idx].no;
		
		//***
		if (mp_no == 1 || mp_no == 25143) {
			set_minor_planet_orientation(mp_no, ds.JD, 0);
		}


		if (mp_no == 25143 && idxItokawaModel >= 0) {
			glModel3D *mdl = get_3DModel(idxItokawaModel);
			double R;
			
			R = Rx * mdl->get_scale();

			glScalef(R, R, R);
			setup_drawing_small_object(idx);
			mdl->draw();
		}
		else {
			glScalef(Rx, Ry, Rz);
			setup_drawing_small_object(idx);
			gl_draw_LOD_sphere();
		}

	glPopMatrix();
}


void
draw_a_small_object_point(const mtk_draw_struct& ds, int idx, const vector3d& rpos, float alpha)
{
	double R2;
	R2 = norm2(rpos);
	if (R2 < 1e-100)  return;

	const orbital_elements& oe = vSmallObjects[idx].oe;

	// V̂̈ʒu
	vector3d  pos = double(_1AU_pc) * oe.get_pos();


	float sz = 1.5f / (ds.rad_per_pixel / 0.001);
	if (sz < 1) {
		sz = 1;
	}


	// xvZ
	const vector3f  lv_e = normalize(get_sun_position()-pos);		// vf̌xNg
	const vector3f  nv = normalize(ds.camera.get_pos() - pos);
	float dp = dot_product(lv_e, nv);
	float th = acosf(dp);
	float frac = 1.0 - th / M_PI;

	glEnable(GL_BLEND);

	// F̌
	float lv = 0.5 * frac + 0.5;
	glColor4f(lv, lv*0.95, lv*0.95, alpha);


	if (true) {
		// QUAD
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);

		vector3f eR, eU;
		if (ds.bRotateBillboard) {
			const small_object& sm = get_small_object(idx);
			eR = sm.eR;
			eU = sm.eU;
		}
		else {
			eR = ds.eRf;
			eU = ds.eUf;
		}

		draw_point(ds, 0.0012, float(sqrt(R2)), rpos, eR, eU);

	} else {
		glPointSize(sz);
		glBegin(GL_POINTS);
			glVertex3dv( rpos.get_pointer() );
		glEnd();
	}
}





//------------
//  Hit-test
//------------
bool
hit_test_small_object(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const vector3d& nv)
{
	vector3d  _rpos, _nv;
	_rpos = rpos;
	_nv = nv;


	// G
	double Rx, Ry, Rz;
	get_small_object_size(idx, Rx, Ry, Rz);
	if (Rx <= 0.0)  return  false;
	Rx *= _1km_pc;
	Ry *= _1km_pc;
	Rz *= _1km_pc;

	double fy, fz;
	fy = Ry / Rx;
	fz = Rz / Rx;
	_rpos.y /= fy;
	_rpos.z /= fz;
	_nv.y /= fy;
	_nv.z /= fz;
	_nv = normalize(_nv);

	// _
	double R, b, c, D, t;
	R = Rx;
	b = dot_product(_rpos, _nv);
	c = _rpos.norm2() - R*R;
	D = b*b - c;

	if (D <= 0.0)  return false;
	t = -b -sqrt(D);
	if (t < 0.0)  return  false;

	return  true;
}



//--------
//  Info
//--------
void
make_object_info_small_object(const mtk_draw_struct& ds, int idx, std::vector<std::string>& text)
{
	const small_object&  smo = vSmallObjects[idx];
	bool bEnglish = (ds.lang == LANG_ENG);

	char type_name[256], str[256];


	if (bEnglish) {
		if (smo.bDwarfPlanet)						{ strcpy(type_name, "Dwarf Planet"); }
		else if (smo.type == SMALLOBJ_TYPE_TNOS)	{ strcpy(type_name, "Trans-Neptunian Object"); }
		else										{ strcpy(type_name, "Minor Planet"); }
	} else {
		if (smo.bDwarfPlanet)						{ strcpy(type_name, "f"); }
		else if (smo.type == SMALLOBJ_TYPE_TNOS)	{ strcpy(type_name, "znOV"); }
		else										{ strcpy(type_name, "f"); }
	}



	//---  Name  ---
	if (smo.label_obj_idx == -1) {
		text.push_back(type_name);
	} else {
		int label_idx = get_object( smo.label_obj_idx ).get_index();
		const char *name = get_label_text(label_idx, bEnglish);
		text.push_back( name );
	}


	//---  Type  ---
	if (bEnglish)	{ sprintf(str, "Type: %s", type_name); }
	else			{ sprintf(str, "ށF%s", type_name); }
	text.push_back(str);


	//---  Number  ---
	if (smo.no != -1) {
		if (bEnglish) {
			sprintf(str, "Minor Planet Number: %d", smo.no);
		} else {
			sprintf(str, "fԍF%d", smo.no);
		}
		text.push_back(str);
	}
}





//------------------
//  visible radius
//------------------
double
calc_visible_radius_small_object(int idx)
{
	return  double(_1AU_pc) * vSmallObjects[idx].oe.get_a();
}



////////////////////////////////////////
/// References
//
// [1] Archinal et al., 2011, Celestial Mechanics and Dynamical Astronomy, 109, 101-135
